<?php


namespace App\HttpController\Google;

use App\Key\K;
use App\Tools\Encoder;
use App\Tools\GSession;
use App\Tools\System;
use EasySwoole\Component\TableManager;
use EasySwoole\EasySwoole\ServerManager;
use Swoole\Coroutine;

class Index extends Base
{

    protected function onRequest(?string $action): ?bool
    {
        parent::onRequest($action);
        $config = System::config();
        $serverId=str_replace(['history'],[''],$config[K::KEY_SERVER_ID]);
        $serverId=intval($serverId);
        if ($serverId<1000){
            $error400 = TableManager::getInstance()->get(K::TABLE_IP_400)->get($this->ip);
            if ($error400['number'] > 0) {
                $this->setStatusCode(500);
                return false;
            }
        }
        $platform = $this->request()->getHeader('cf');
        if (strtolower(current($platform)) === 'mobile' && !$config[K::KEY_HISTORY_MOBILE_ENABLE]) {
            $this->setStatusCode(404);
            return false;
        }
        if (!$config[K::KEY_GOOGLE_HISTORY_ENABLE]) {
            $this->setStatusCode(403);
            return false;
        }
        $limit = intval($config[K::KEY_EVERY_SECOND_LIMIT]);
        if (!is_numeric($limit) || $limit > 300 || $limit < 1) $limit = 30;
        if ($this->request()->getMethod() === 'GET' && $this->request()->getUri()->getPath() == '/google/index/tile'){
            if ($config[K::KEY_REQUEST_WAITING_ENABLE]){
                TableManager::getInstance()->get(K::TABLE_SYSTEM_CONFIG)->incr(K::KEY_EVERY_SECOND_REQUEST_ALL, 'number');
                $waitingTable=TableManager::getInstance()->get(K::TABLE_QUERY_WAITING);
                if ($waitingTable->count()>$limit*2) {
                    $this->setStatusCode(500);
                    return false;
                }
                $cid=Coroutine::getCid();
                $workerId=ServerManager::getInstance()->getSwooleServer()->worker_id;
                $pauseKey=$cid.'-'.$workerId;
//            echo 'pause key ',$pauseKey,PHP_EOL;
                $waitingTable->set($pauseKey,['time'=>microtime(true),'cid'=>$cid,'worker_id'=>$workerId,'code'=>0]);
                Coroutine::suspend();
                $resumeData=$waitingTable->get($pauseKey);
                $waitingTable->del($pauseKey);
                if ($resumeData['code']!=200){
                    $this->setStatusCode(500);
                    return  false;
                }
            }else{
                if (TableManager::getInstance()->get(K::TABLE_SYSTEM_CONFIG)->incr(K::KEY_EVERY_SECOND_REQUEST_ALL,'number')>$limit){
                    $this->setStatusCode(500);
                    return  false;
                }
            }
        }
        return  true;
    }

    public function ip()
    {
        $this->response()->write($this->ip);
    }

    public function db()
    {
        $originRequest = $this->request()->getSwooleRequest();
        $url = 'https://kh.google.com/dbRoot.v5?' . $originRequest->server['query_string'];
        System::log($url, 'db root');
        //echo $this->request()->getQueryParam('db'),PHP_EOL;
        if ($this->request()->getQueryParam('ct')) {
            if ($this->request()->getQueryParam('db')) {
                //历史影像的数据
                $name = date('Y-m-d') . '_history';
            } else {
                $name = date('Y-m-d') . '_today';
            }
            $cachePath = System::config(K::KEY_CACHE_PATH);
            $cachePath = $cachePath . '/db';
            if (!is_dir($cachePath) && !mkdir($cachePath)) System::log('mkdir db cache dir failed');
            $data = file_get_contents($cachePath . '/' . $name);
            if ($data) {
                $this->response()->getSwooleResponse()->header('content-type', 'application/octet-stream');
                $this->response()->getSwooleResponse()->write($data);
                return;
            }
        }
        try {
            $saber = System::getSaber();
            $curlData = $saber->get($url, ['cookies' => $this->getCookie(), 'headers' => $this->header()]);
            $responseHeader = $curlData->getHeader('content-type');
            if ($responseHeader) $this->response()->getSwooleResponse()->header('content-type', $responseHeader);
            $this->response()->getSwooleResponse()->setStatusCode($curlData->getStatusCode());
            $data = $curlData->getBody()->getContents();
            if ($name) {
                file_put_contents($cachePath . '/' . $name, $data);
            }
            $this->response()->getSwooleResponse()->write($data);
        } catch (\Throwable $exception) {
            System::log($exception->getMessage(), 'db root error');
            $this->setStatusCode(403);
            $this->response()->getSwooleResponse()->write($exception->getMessage());
        }
    }

    public function tile()
    {
        $originRequest = $this->request()->getSwooleRequest();
        $queryString=$originRequest->server['query_string'];
        if (substr($queryString, -1) === '=') $queryString = substr($queryString, 0, -1);
//        $url = 'https://kh.google.com/flatfile?' . str_replace(['_'],['+'],$queryString);
        $url = 'https://kh.google.com/flatfile?' .$queryString;
//        echo $url,PHP_EOL;
//        if (mt_rand(0,10)<1) System::log($this->header(),'header');
//        file_put_contents(__DIR__.'/b.url',$url.PHP_EOL,FILE_APPEND);
        $header = $this->header();
        $session = GSession::getSession('second');
        if (!$session) {
            $this->setStatusCode(403);
            return false;
        }
        $cookie = $this->getCookie();
        $length = unpack('C', substr($session, 7, 1));
        $session = substr($session, 8, current($length));
        $cookie['SessionId'] = $session;
        try {
            $saber = System::getSaber();
            $data = $saber->get($url, ['timeout' => 5, 'cookies' => $cookie, 'headers' => $header]);
//            $data = SaberGM::get($url, ['timeout' => 3, 'cookies' => $cookie, 'headers' => $header]);
            $this->header('content-type','application/octet-stream');
            $content = $data->getBody()->getContents();
            //统计每天的上传量
            TableManager::getInstance()->get(K::TABLE_SYSTEM_CONFIG)->incr(K::KEY_EVERY_DAY_UPLOAD, 'number', strlen($content)*1.1);
            //每秒请求正常的次数
            TableManager::getInstance()->get(K::TABLE_SYSTEM_CONFIG)->incr(K::KEY_EVERY_SECOND_REQUEST_OK, 'number');
            TableManager::getInstance()->get(K::TABLE_SYSTEM_CONFIG)->delete(K::KEY_TILE_ERROR_TIMES);
            $this->response()->getSwooleResponse()->write($content);
//            echo 'this is OK session',PHP_EOL,$session,PHP_EOL;
        } catch (\Throwable $exception) {
            $errorCode = $exception->getCode();
//            echo 'this is no ok code:',$errorCode,PHP_EOL,$session,PHP_EOL;
            switch ($errorCode) {
                case 403:
                    $secondTable = TableManager::getInstance()->get(K::TABLE_SECOND_CACHE);
                    $delKey = md5($cookie['SessionId']);
//                    $sessionData = $secondTable->get($delKey);
//                    if ($sessionData) {
//                        System::log(time() - $sessionData['time'], 'current live time');
//                    }
                    if ($secondTable->del($delKey)) {
                        System::log($delKey, 'ip ' . $this->ip . ' delete session ');
                    }
                    break;
                case 404:
                    break;
                case -1:
                case -2:
                case -3:
                    //这里是超时的错误，比如代理无法访问之类的，或者是流量使用完了，
                    $errorCode = 500;
                    if (TableManager::getInstance()->get(K::TABLE_SYSTEM_CONFIG)->incr(K::KEY_TILE_ERROR_TIMES, 'number') > 200) {
                        //如果出错了200次，有可能是代理的流量使用完了，那么也要暂时服务器，等待判断
                        System::config(K::KEY_GOOGLE_HISTORY_ENABLE, '');
                        System::log('', 'too many tile errors stop server');
                    }
                    break;
                case 400:
                    $error400 = TableManager::getInstance()->get(K::TABLE_IP_400)->get($this->ip);
                    if (!$error400) $error400 = ['number' => 0];
                    $error400['number']++;
                    $error400['time'] = time();
                    TableManager::getInstance()->get(K::TABLE_IP_400)->set($this->ip, $error400);
                    System::log($errorCode . '    ' . $url, 'code 400   ' . $this->ip);
//                    System::log($this->request()->getHeaders(),'error code 400 header');
                    $errorCode = 500;
                    break;
                default:
                    //如果是未知的错误码，那么返回500错误
                    System::log($errorCode . '    ' . $url, 'unknown tile error code ip   ' . $this->ip);
                    $errorCode = 500;
                    if (TableManager::getInstance()->get(K::TABLE_SYSTEM_CONFIG)->incr(K::KEY_TILE_ERROR_TIMES, 'number') > 200) {
                        //如果出错了5000次，有可能是代理的流量使用完了，那么也要暂时服务器，等待判断
                        System::config(K::KEY_GOOGLE_HISTORY_ENABLE, '');
                        System::log('', 'too many else errors stop server');
                    }
            }
            $this->setStatusCode($errorCode);
            $this->response()->getSwooleResponse()->write($exception->getCode().' '.$exception->getMessage());
        }
    }

    public function auth()
    {
        $cookie = $this->getCookie();
        if (isset($cookie['SessionId'])) {
            //因为session 的删除不是由客户端决定的，是服务器来决定的，这里不再删除
//            $delKey = md5($cookie['SessionId']);
//            TableManager::getInstance()->get(K::TABLE_SECOND_CACHE)->del($delKey);
        }
        //$this->savePostData();
        $type=$this->request()->getQueryParam('ct')?'second':'first';
        if ($type=='first') {
            //这个是第一次POST请求的内容,那么要判断是不是要放行
            $pass = System::config(K::KEY_FIRST_POST_PASS);
            if ($pass > mt_rand(1, 100)) {
                //这里的的数据是直接请求的
                $saber = System::getSaber();
                try {
                    $default = GSession::getDefaultQueryData('first');
                    $this->response()->getSwooleResponse()->header('content-type', 'application/octet-stream');
                    $first_data = $saber->post($default['url'], $this->request()->getSwooleRequest()->getContent(), ['headers' => $default['headers']]);
                    $first_result = $first_data->getBody()->getContents();
                    System::log(base64_encode($first_result), 'ip   ' . $this->ip . '  first data pass');
                    $this->response()->getSwooleResponse()->write($first_result);
                } catch (\Throwable $exception) {
                    System::log($exception->getMessage(), 'first data pass error');
                    $this->response()->withStatus($exception->getCode());
                    $this->response()->getSwooleResponse()->setStatusCode($exception->getCode());
                    $this->response()->getSwooleResponse()->status($exception->getCode());
                    $this->response()->getSwooleResponse()->write($exception->getMessage());
                }
                return false;
            }
        }
        $data = GSession::getSession($type);
        if (!$data) {
            Coroutine::sleep(1);
            $data = GSession::getSession($type);
        }
        if ($data) {
//            print_r($data);
            //这里的header固定死了
            $this->response()->getSwooleResponse()->header('content-type', 'application/octet-stream');
            if ($this->request()->getQueryParam('ct')) {
                $encoder = new Encoder();
                return $this->response()->getSwooleResponse()->write($encoder->encode($data));
            } else {
                System::log(base64_encode($data), 'return first base64 ip  ' . $this->ip);
                return $this->response()->getSwooleResponse()->write($data);
//                $this->response()->write($data);
            }
        } else {
            $this->response()->withStatus(403);
        }
    }

    private function savePostData()
    {
        $postData = $this->request()->getSwooleRequest()->getContent();
        if (strlen($postData) != 47 && strlen($postData) != 49) {
            System::log('empty post data ' . date('Y-m-d H:i:s'));
            return;
        }
        $table = TableManager::getInstance()->get(K::TABLE_QUERY_LIST);
        $header = $this->header();
        if ($this->request()->getQueryParam('ct')) {
            TableManager::getInstance()->get(K::TABLE_SYSTEM_CONFIG)->incr(K::KEY_EVERY_DAY_POST2, 'number');
        } else {
            TableManager::getInstance()->get(K::TABLE_SYSTEM_CONFIG)->incr(K::KEY_EVERY_DAY_POST1, 'number');
        }
        if ($table->count() < 100) {
            $cacheData = $table->get($this->ip);
            $cachePath = System::config(K::KEY_CACHE_PATH);
            if ($this->request()->getQueryParam('ct')) {
                $cacheData['second_time'] = time();
                $cacheData['second_data'] = base64_encode($postData);
                $cacheData['second_header'] = json_encode($header, JSON_UNESCAPED_SLASHES);
                file_put_contents($cachePath . '/post2-' . date('Y-m-d'), json_encode(['ip' => $this->ip, 'data' => $cacheData]) . PHP_EOL, FILE_APPEND);
            } else {
                $cacheData['first_time'] = time();
                $cacheData['first_data'] = base64_encode($postData);
                $cacheData['first_header'] = json_encode($header, JSON_UNESCAPED_SLASHES);
                file_put_contents($cachePath . '/post1-' . date('Y-m-d'), json_encode(['ip' => $this->ip, 'data' => $cacheData]) . PHP_EOL, FILE_APPEND);
            }
            if (strpos($cacheData['second_data'], 'AQAAAAM3vfBl6') === 0) {
                //这个是默认的手机端的数据,直接不使用
                return;
            }
            $table->set($this->ip, $cacheData);
        } else {
            $delKey = [];
            foreach ($table as $k => $v) {
                if (!$v['second_data']||mt_rand(0,100)>=50) {
                    $delKey[] = $k;
                }
            }
            foreach ($delKey as $v) {
                $table->del($v);
            }
        }
    }
}
