<?php
namespace App\Tools;
class Encoder
{
    const ROUND_NUMBER_BYTES = 4;
    /**
     * Encode content, Internal of Bigemap.
     *
     * @param string|array $content String or bytes array content
     */
    public function encode($content)
    {
        $content_bytes = array();
        if (is_array($content)) {
            $content_bytes = $content;
        } else if (is_string($content)) {
            $content_bytes = BMUtils::str2bytes($content);
        }

        // Get key seed.
        #$time = explode(" ", microtime());

        #
        # key_bytes[0-7] is always 0, when key_seed len is 8.
        #
        # $temp = substr($time[1], 2) + substr($time[0], 2);
        # if (strlen($temp) > 8) {
        #     $temp = substr($temp, 0, 8);
        # }
        #
        # $key_seed = $time[1] + substr($time[0], 2);
        #

        $key_seed_bytes = array();
        for ($i = 0; $i < self::ROUND_NUMBER_BYTES; $i++) {
            $key_seed_bytes[$i] = mt_rand(127, 255);
        }

        // generate key
        $key_bytes = self::_generateKey($key_seed_bytes);

        // crypt tata
        self::_crypt($key_bytes, $content_bytes);

        $crypted = BMUtils::bytes2str($key_seed_bytes, $content_bytes);

        return $crypted;
    }

    /**
     * Decode content, Internal of Bigemap.
     *
     * @param string|array $crypted_content String or bytes array of crypted content.
     * @param int $pos Start position in crypted content, default 0.
     */
    public function decode($crypted_content, $pos = 0)
    {
        $content_bytes = array();
        if (is_array($crypted_content)) {
            $content_bytes = $crypted_content;
        } else if (is_string($crypted_content)) {
            $content_bytes = BMUtils::str2bytes($crypted_content);
        }

        if (count($content_bytes) <= $pos + self::ROUND_NUMBER_BYTES) {
            return "";
        }

        $key_seed = array_slice($content_bytes, $pos, self::ROUND_NUMBER_BYTES);
        $key_bytes = self::_generateKey($key_seed);

        $content_pos = $pos + self::ROUND_NUMBER_BYTES;
        $content_bytes = array_slice($content_bytes, $content_pos);

        self::_decrypt($key_bytes, $content_bytes);

        $uncrypted = BMUtils::bytes2str($content_bytes);
        return $uncrypted;
    }

    /**
     * Generate key bytes array.
     *
     * @param string|int|array $key_seed Seed of Key. byte array(4), int or string(int).
     */
    private static function _generateKey($key_seed)
    {
        if (is_array($key_seed) && count($key_seed) == 4) {
            $key_bytes = $key_seed;
        } else if (is_int($key_seed)) {
            $key_bytes = BMUtils::int2bytes($key_seed, false);
        } else if (is_string($key_seed)) {
            $key_bytes = BMUtils::int2bytes(intval($key_seed), false);
        }

        $key = array();
        for ($i = 0; $i < 4; $i++) {
            for ($j = 0; $j < 8; $j++) {
                if ($key_bytes[$i] >= 128)
                    $key[$i * 8 + $j] = (($key_bytes[$i] >> $j) - (256 >> $j));
                else
                    $key[$i * 8 + $j] = ($key_bytes[$i] >> $j);
            }
        }

        return $key;
    }

    /**
     * crypt data.
     */
    private static function _crypt($key_bytes_array, &$content_bytes_array)
    {
        $key_len = count($key_bytes_array);
        $content_len = count($content_bytes_array);
        for ($i = 0; $i < $content_len; $i++) {
            $content_bytes_array[$i] = $content_bytes_array[$i] + $key_bytes_array[$i % $key_len];
        }
    }

    /**
     * decrypt data.
     */
    private static function _decrypt($key_bytes_array, &$content_bytes_array)
    {
        $key_len = count($key_bytes_array);
        $content_len = count($content_bytes_array);
        for ($i = 0; $i < $content_len; $i++) {
            $content_bytes_array[$i] = $content_bytes_array[$i] - $key_bytes_array[$i % $key_len];
        }
    }
}
