<?php
namespace App\Tools;
use App\Key\K;
use EasySwoole\Component\Singleton;

class ProxyManager{
    use Singleton;

    /**
     * @var \SQLite3
     */
    private $connect;
    public function __construct()
    {
        try {
            $cache=System::config(K::KEY_CACHE_PATH).'/proxy.db';
            $this->connect=new \SQLite3($cache);
            $this->connect->busyTimeout(5000);
        }catch (\Throwable $exception){
            System::log($exception->getMessage(),'proxy error');
        }
    }
    public function getAllServer(){
        $query=$this->connect->query('select * from subscribe order  by id desc');
        $result=[];
        while ($t=$query->fetchArray(SQLITE3_ASSOC)){
            $t['children']=$this->select('server_list',['pid'=>$t['id']]);
//            $t['list']=$t['children'];
            //添加一个当前的使用量的值;
            $t['in_use']=0;
            $t['_all_']=count($t['children']);
            $t['_unique_']=count(array_unique(array_column($t['children'],'ip')));
            $t['_forbidden_']=[];
            $t['_never_']=[];
            foreach ($t['children'] as $v){
                //计算一下有多少IP被封,多少个代理从来没有被使用过
                if ($v['status']==100) $t['_forbidden_'][]=$v['ip'];
                if (!$v['ip']) $t['_never_'][]=$v;
            }
            $t['_forbidden_']=count(array_unique($t['_forbidden_']));
            $t['_never_']=count($t['_never_']);
            $result[]=$t;
        }
        //然后计算一下，当前的订阅的使用人数
        $allServerData = System::serverData();
        $usedUnique = [];
        foreach ($allServerData as $v) {
            if ($v[K::KEY_PROXY_HISTORY_ENABLE] && $v[K::KEY_PROXY_CURRENT][0]) {
                $usedUnique[] = $v[K::KEY_PROXY_CURRENT][0] . '_' . $v[K::KEY_SERVER_ID];
            }
        }
        foreach ($result as $k => $v) {
            $used = [];
            foreach ($v['children'] as $vv) {
                if ($vv['status'] == 2 && $vv['server_id'] && in_array($vv['pid'] . '_' . $vv['server_id'], $usedUnique)) {
                    $used[$vv['server_id']] = 1;
                    $result[$k]['in_use'] = count($used);
                }
            }
        }
        return $result;
    }
    public function add($table,$data){
        $fields=$this->getTableFields($table);
        foreach ($data as $k=>$v){
            if (!in_array($k,$fields)) unset($data[$k]);
        }
        $values=array_map(function ($v){
                return "'".$v."'";
        },array_values($data));
        $sql=str_replace(['__table__','__key__','__values__'],[$table,'`'.implode('` , `',array_keys($data)).'`',implode(' , ',$values)],'insert into __table__ (__key__) VALUES (__values__)');
        if (!$this->connect->query($sql)){
            System::log($this->connect->lastErrorMsg(),$sql);
            return false;
        }
        return  $this->connect->lastInsertRowID();
    }

    public function delete($table,$where){
        $w=[];
        foreach ($where as $k=>$v){
            $w[]='`'.$k. '` ='.' \''.$v.'\'';
        }
        $sql=str_replace(['__table__','__where__'],[$table,implode(' , ',$w)],'delete from __table__ where __where__');
        if (!$this->connect->query($sql)){
            System::log($this->connect->lastErrorMsg(),$sql);
            return false;
        }
        return  true;
    }
    public function save($table,$where,$data){
        foreach ($where as $k=>$v){
            $w[]=$k. '='.' \''.$v.'\'';
        }
        $fields=$this->getTableFields($table);
        foreach ($data as $k=>$v){
            if (!in_array($k,$fields)) continue;
            $d[]='`'.$k.'` ='.' \''.$v.'\'';
        }
        $sql=str_replace(['__table__','__set__','__where__'],[$table,implode(' , ',$d),implode(' ADN ',$w)],'update __table__ set __set__ WHERE __where__');
        if (!$this->connect->query($sql)){
            System::log($this->connect->lastErrorMsg(),$sql);
            return false;
        }
        return  true;
    }

    public function select($table,$where,$arrayKey=''){
        foreach ($where as $k=>$v){
            $w[]=$k. '='.' \''.$v.'\'';
        }
        $sql=str_replace(['__table__','__where__'],[$table,implode(' ADN ',$w)],'select * from __table__ where __where__');
        $result=$this->connect->query($sql);
        if (!$result){
            System::log($this->connect->lastErrorMsg(),$sql);
            return false;
        }
        $d=[];
        while ($t=$result->fetchArray(SQLITE3_ASSOC)){
            if ($arrayKey){
                $d[$t[$arrayKey]]=$t;
            }else{
                $d[]=$t;
            }
        }
        return  $d;
    }

    public function find($table,$where){
        $data=$this->select($table,$where);
        $data=current($data);
        return $data??[];
    }
    public function reSubscribeLink($id){
        //如果重新订阅地址
        $data=$this->find('subscribe',['id'=>$id]);
        if (!$data['url']){
            System::log('no resubscribe url find');
            return  false;
        }
        $serverInfo = Proxy::parseUrl($data['url']);
        if (!$serverInfo) {
            System::log('resubscribe failed');
            return  false;
        }
        //更新一次这个订阅的最后订阅更新时间,因为会默认每隔三天更新订阅，这里设置一个随机的时间，防止平台来检测
        $this->save('subscribe',['id'=>$id],['resubscribe_time'=>time()+mt_rand(3600*8,3600*24)]);
        $oldData = $this->select('server_list', ['pid' => $data['id']], 'key');
        //求出已经是过期订阅的服务器
        $removeServer = array_diff_key($oldData, $serverInfo);
        foreach ($removeServer as $v) {
            $this->delete('server_list', ['id' => $v['id']]);
        }
        foreach ($serverInfo as $k => $v) {
            if (!$oldData[$k]) {
                //本地没有保存这个服务器，添加新的
                $addData = ['key' => $k, 'pid' => $data['id'], 'info' => json_encode($v, JSON_UNESCAPED_UNICODE | JSON_UNESCAPED_SLASHES)];
                $this->add('server_list', $addData);
            } else {
                //更新数据
                $oldData[$k]['info'] = json_encode($v, JSON_UNESCAPED_UNICODE | JSON_UNESCAPED_SLASHES);
                $this->save('server_list', ['id' => $oldData[$k]['id']], $oldData[$k]);
            }
        }
        return  true;
    }
    public function getTableFields($table){
        $sql='PRAGMA table_info('.$table.')';
        $result=$this->connect->query($sql);
        while ($t=$result->fetchArray(SQLITE3_ASSOC)){
            $d[]=$t;
        }
        return array_column($d,'name');
    }
}